#include "MQTTTelemetry.h"

#include <ArduinoJson.h>

#include "../../Static.h"
#include "../../Config.h"
#include "../../Platform.h"

struct MQTTDevice {
  const String id;
  const String name;
  const String model;
  const String softwareVersion;
  const String manufacturer;
  const String availabilityTopic;

  void toJson(const JsonObject& json) const {
    json["name"] = name;
    json["mdl"] = model;
    json["sw"] = softwareVersion;
    json["mf"] = manufacturer;
    json["ids"][0] = id;
  }
};

const String availTopic = String("renderbug/") + Platform::deviceID() + "/availability";

const MQTTDevice Device{
  Platform::deviceID(),
  Platform::deviceName(),
  Platform::model(),
#ifdef BOARD_ESP8266
  ESP.getSketchMD5(),
#else
  "",
#endif
  "Phong Robotics",
  availTopic
};

struct MQTTEntity {
  const MQTTDevice& device;
  String name;
  String entityId;
  String rootTopic;

  MQTTEntity(const String& domain, const MQTTDevice& device, const String& name) : device(device), name(Platform::deviceName() + " " + name) {
    entityId = String(device.id) + "-" + name;
    rootTopic = String("homeassistant/") + domain + String("/renderbug/") + entityId;
  }

  String configTopic() const {
    return rootTopic + "/config";
  }

  String commandTopic() const {
    return rootTopic + "/set";
  }

  String heartbeatTopic() const {
    return String("renderbug/") + Device.id + "/heartbeat";
  }

  String statTopic() const {
    return rootTopic + "/state";
  }

  bool isCommandTopic(const char* topic) const {
    if (strncmp(topic, rootTopic.c_str(), rootTopic.length()) == 0) {
      return strncmp(&topic[rootTopic.length()], "/set", sizeof("/set")) == 0;
    }
    return false;
  }

  void toJson(JsonDocument& jsonBuf, bool isInteractive = true) const {
    jsonBuf["~"] = rootTopic.c_str();
    jsonBuf["name"] = name;
    jsonBuf["unique_id"] = entityId;
    if (isInteractive) {
      jsonBuf["cmd_t"] = "~/set";
      jsonBuf["ret"] = true;
      jsonBuf["schema"] = "json";
    } else {
    }
    jsonBuf["stat_t"] = "~/state";
    jsonBuf["json_attr_t"] = heartbeatTopic();
    jsonBuf["avty_t"] = device.availabilityTopic;
    device.toJson(jsonBuf.createNestedObject("dev"));
  }
};

const MQTTEntity Lightswitch {
  "light", Device, "lightswitch"
};

const MQTTEntity flashlightSwitch {
  "switch", Device, "flashlight"
};

const MQTTEntity FPSSensor {
  "sensor", Device, "fps"
};

MQTTTelemetry::MQTTTelemetry() : BufferedInputSource("MQTT"),
    m_mqtt(m_wifi),
    m_logPrinter(this)
{
    m_debugTopic = String("renderbug/") + Platform::deviceID();
}

void
MQTTTelemetry::handleEventOnline(const InputEvent& evt)
{
  if (!m_mqtt.connected()) {
    Log.notice("Connecting to MQTT as %s on %s...", Platform::deviceID(), Device.availabilityTopic.c_str());
    if (m_mqtt.connect(Platform::deviceID(), NULL, NULL, Device.availabilityTopic.c_str(), 0, true, "offline")) {
      Log.notice("Connected to MQTT");
      m_needHeartbeat = true;

      StaticJsonDocument<1024> configJson;

      Lightswitch.toJson(configJson);

      int i = 0;
      for(const Sequencer::Scene& scene : m_sequencer->scenes()) {
        configJson["fx_list"][i++] = scene.name;
      }
      configJson["brightness"] = true;
      configJson["rgb"] = true;

      char buf[1024];
      serializeJson(configJson, buf, sizeof(buf));

      Log.verbose("Publish %s %s", Lightswitch.configTopic().c_str(), buf);
      m_mqtt.publish(Lightswitch.configTopic().c_str(), (uint8_t*)buf, strlen(buf), true);
      m_mqtt.subscribe(Lightswitch.commandTopic().c_str());

      configJson.clear();
      flashlightSwitch.toJson(configJson, false);
      configJson["cmd_t"] = "~/set";
      configJson["ret"] = true;
      serializeJson(configJson, buf, sizeof(buf));
      m_mqtt.publish(flashlightSwitch.configTopic().c_str(), (uint8_t*)buf, strlen(buf), true);
      m_mqtt.subscribe(flashlightSwitch.commandTopic().c_str());

      configJson.clear();
      FPSSensor.toJson(configJson, false);
      configJson["unit_of_meas"] = "Frames/s";
      serializeJson(configJson, buf, sizeof(buf));

      Log.verbose("Publish %s %s", FPSSensor.configTopic().c_str(), buf);
      m_mqtt.publish(FPSSensor.configTopic().c_str(), (uint8_t*)buf, strlen(buf), true);
      m_mqtt.subscribe(FPSSensor.commandTopic().c_str());

#ifdef BOARD_ESP8266
      struct rst_info resetInfo = *ESP.getResetInfoPtr();
      if (resetInfo.reason != 0) {
        char buff[200];
        sprintf(&buff[0], "Fatal exception:%d flag:%d (%s) epc1:0x%08x epc2:0x%08x epc3:0x%08x excvaddr:0x%08x depc:0x%08x", resetInfo.exccause, resetInfo.reason, (resetInfo.reason == 0 ? "DEFAULT" : resetInfo.reason == 1 ? "WDT" : resetInfo.reason == 2 ? "EXCEPTION" : resetInfo.reason == 3 ? "SOFT_WDT" : resetInfo.reason == 4 ? "SOFT_RESTART" : resetInfo.reason == 5 ? "DEEP_SLEEP_AWAKE" : resetInfo.reason == 6 ? "EXT_SYS_RST" : "???"), resetInfo.epc1, resetInfo.epc2, resetInfo.epc3, resetInfo.excvaddr, resetInfo.depc); 
        Log.warning("Previous crash detected! %s", buff);
      }
#endif
    } else {
      Log.warning("Could not connect to MQTT");
    }
  } else {
    String statTopic = Lightswitch.statTopic();
    if (evt.intent == InputEvent::StopThing && String(evt.asString()) == "Flashlight") {
      String flashlightStatTopic = flashlightSwitch.statTopic();
      m_mqtt.publish(flashlightStatTopic.c_str(), "OFF");
    } else if (evt.intent == InputEvent::StartThing && String(evt.asString()) == "Flashlight") {
      String flashlightStatTopic = flashlightSwitch.statTopic();
      m_mqtt.publish(flashlightStatTopic.c_str(), "ON");
    } else if (evt.intent == InputEvent::SetPower) {
      StaticJsonDocument<256> doc;
      char buf[256];
      m_isOn = evt.asInt() ? true : false;
      doc["state"] = m_isOn ? "ON" : "OFF";
      serializeJson(doc, buf, sizeof(buf));
      m_mqtt.publish(statTopic.c_str(), buf);
    } else if (evt.intent == InputEvent::SetBrightness) {
      StaticJsonDocument<256> doc;
      char buf[256];
      doc["brightness"] = evt.asInt();
      doc["state"] = m_isOn ? "ON" : "OFF";
      serializeJson(doc, buf, sizeof(buf));
      m_mqtt.publish(statTopic.c_str(), buf);
    } else if (evt.intent == InputEvent::SetColor) {
      StaticJsonDocument<256> doc;
      char buf[256];
      CRGB color = evt.asRGB();
      doc["color"]["r"] = color.r;
      doc["color"]["g"] = color.g;
      doc["color"]["b"] = color.b;
      doc["state"] = m_isOn ? "ON" : "OFF";
      serializeJson(doc, buf, sizeof(buf));
      m_mqtt.publish(statTopic.c_str(), buf);
    } else if (evt.intent == InputEvent::SetPattern) {
      StaticJsonDocument<256> doc;
      char buf[256];
      doc["effect"] = evt.asString();
      doc["state"] = m_isOn ? "ON" : "OFF";
      serializeJson(doc, buf, sizeof(buf));
      m_mqtt.publish(statTopic.c_str(), buf);
    } else if (evt.intent == InputEvent::FirmwareUpdate) {
      String updateTopic = m_debugTopic + "/firmware";
      m_mqtt.publish(updateTopic.c_str(), "firmware update!");
    }
  }
}

void
MQTTTelemetry::loop()
{
  BufferedInputSource::loop();
  OnlineTaskMixin::loop();
}

void
MQTTTelemetry::onOnline()
{
  const IPAddress server(10, 0, 0, 2);

  m_needHeartbeat = true;
  m_mqtt.setServer(server, 1883);
  m_mqtt.setBufferSize(1024);
  m_mqtt.setCallback(&MQTTTelemetry::s_callback);
}

void
MQTTTelemetry::onOffline()
{
  m_mqtt.disconnect();
}

void
MQTTTelemetry::loopOnline()
{
  m_mqtt.loop();
  EVERY_N_SECONDS(10) {
    m_needHeartbeat = true;
  }
  if (m_needHeartbeat) {
    char buf[512];
    StaticJsonDocument<512> response;
    response["device_id"] = Platform::deviceID();
    response["sketch_version"] = ESP.getSketchMD5();
    response["os_version"] = ESP.getSdkVersion();
    response["localip"] = WiFi.localIP().toString();
    response["pixelCount"] = Static<ConfigService>::instance()->coordMap()->pixelCount;
    response["startPixel"] = Static<ConfigService>::instance()->coordMap()->startPixel;
    response["RSSI"] = WiFi.RSSI();
    response["free_ram"] = ESP.getFreeHeap();
    response["fps"] = FastLED.getFPS();
    serializeJson(response, buf, sizeof(buf));
    String availTopic = m_rootTopic + "/available";
    m_mqtt.publish(Lightswitch.heartbeatTopic().c_str(), buf);
    m_mqtt.publish(Device.availabilityTopic.c_str(), "online");
    //Log.notice("Heartbeat: %s", buf);

    String fpsCounter = String(FastLED.getFPS());
    m_mqtt.publish(FPSSensor.statTopic().c_str(), fpsCounter.c_str());

    m_needHeartbeat = false;
  }
}

void
MQTTTelemetry::callback(char* topic, const char* payload)
{
    setEvent(InputEvent::NetworkActivity);
    if (flashlightSwitch.isCommandTopic(topic)) {
      if (!strncmp((char*)payload, "ON", sizeof("ON"))) {
        Log.notice("Turning on flashlight");
        setEvent(InputEvent{InputEvent::SetPower, true});
        setEvent(InputEvent{InputEvent::SetPattern, "Flashlight"});
        setEvent(InputEvent{InputEvent::SetBrightness, 255});
      } else if (!strncmp((char*)payload, "OFF", sizeof("OFF"))) {
        Log.notice("Turning off flashlight");
        setEvent(InputEvent{InputEvent::SetPattern, "Idle"});
      }
    } else if (Lightswitch.isCommandTopic(topic)) {
      StaticJsonDocument<512> doc;
      deserializeJson(doc, payload);

      if (doc.containsKey("state")) {
          if (doc["state"] == "ON") {
            Log.notice("Turning on power");
            setEvent(InputEvent{InputEvent::SetPower, true});
          } else if (doc["state"] == "OFF") {
            Log.notice("Turning off power");
            setEvent(InputEvent{InputEvent::SetPattern, "Idle"});
            setEvent(InputEvent{InputEvent::SetPower, false});
          }
      }

      if (doc.containsKey("start")) {
        strcpy(m_patternBuf, doc["start"].as<const char*>());
        setEvent(InputEvent{InputEvent::StartThing, m_patternBuf});
      }

      if (doc.containsKey("stop")) {
        if (doc["stop"] == name) {
          Log.notice("You can't kill an idea, or stop the MQTT Task via MQTT.");
        } else {
          strcpy(m_patternBuf, doc["stop"].as<const char*>());
          setEvent(InputEvent{InputEvent::StopThing, m_patternBuf});
        }
      }

      if (doc.containsKey("pixelCount")) {
        setEvent(InputEvent{InputEvent::SetDisplayLength, (int)doc["pixelCount"]});
      }

      if (doc.containsKey("startPixel")) {
        setEvent(InputEvent{InputEvent::SetDisplayOffset, (int)doc["startPixel"]});
      }

      if (doc.containsKey("save")) {
        setEvent(InputEvent{InputEvent::SaveConfigurationRequest});
      }

      if (doc.containsKey("restart")) {
#ifdef BOARD_ESP8266
        ESP.wdtDisable();
        ESP.restart();
#endif
      }

      if (doc.containsKey("reconnect")) {
        m_mqtt.disconnect();
      }

      if (doc.containsKey("ping")) {
        m_needHeartbeat = true;
        Log.notice("Queuing up heartbeat");
      }

      if (doc.containsKey("effect")) {
          strcpy(m_patternBuf, doc["effect"].as<const char*>());
          setEvent(InputEvent{InputEvent::SetPattern, m_patternBuf});
      }

      if (doc.containsKey("color")) {
          uint8_t r = doc["color"]["r"];
          uint8_t g = doc["color"]["g"];
          uint8_t b = doc["color"]["b"];
          setEvent(InputEvent{InputEvent::SetColor, CRGB(r, g, b)});
      }

      if (doc.containsKey("brightness")) {
          setEvent(InputEvent{InputEvent::SetBrightness, (int)doc["brightness"]});
      }
    }
}

void
MQTTTelemetry::s_callback(char* topic, byte* payload, unsigned int length)
{
    char topicBuf[128];
    char payloadBuf[512];
    strcpy(topicBuf, topic);
    memcpy(payloadBuf, payload, length);
    payloadBuf[std::min(sizeof(payloadBuf) - 1, length)] = 0;
    Static<MQTTTelemetry>::instance()->callback(topicBuf, payloadBuf);
}

STATIC_ALLOC(MQTTTelemetry);
